from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

TJsonLiteralValue = Union[str, int, float, bool, None]
TJsonKey = Union[str, int]
TJsonStructure = Union[Dict[TJsonKey, "TJson"], List["TJson"], Tuple["TJson", ...]]
TJson = Union[TJsonLiteralValue, TJsonStructure]

TJsonCustomSerializer = Callable[[Any], TJson]

# 因为 Tuple 虽然是 Hashable & builtin & Immutable, 但是不符合 JSON 规范, 所以 Tuple 无法作为 TJsonKey.
# bytes 不必说.


class JSONBackend(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, value: TJson, *, custom_serializers: Optional[Dict[Type, TJsonCustomSerializer]] = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def deserialize(self, value: str) -> TJson:
        raise NotImplementedError()

    def serialize_as_bytes(
        self, value: Any, *, custom_serializers: Optional[Dict[Type, TJsonCustomSerializer]] = None
    ) -> bytes:
        return self.serialize(value, custom_serializers=custom_serializers).encode("utf-8")


from .bootstrap import CURRENT_BACKEND as CURRENT_BACKEND
from .frontend import Json as Json



from abc import ABCMeta, abstractmethod
from datetime import timedelta
from typing import Any, Generic, List, Optional, TypeVar

from launart.service import ExportInterface

D = TypeVar("D")


class Storage(ExportInterface):
    pass


class CacheStorage(Storage, Generic[D], metaclass=ABCMeta):
    @abstractmethod
    async def get(self, key: str, default: Optional[Any] = None) -> D:
        ...

    @abstractmethod
    async def set(self, key: str, value: Any, expire: Optional[timedelta] = None) -> None:
        ...

    @abstractmethod
    async def delete(self, key: str, strict: bool = False) -> None:
        ...

    @abstractmethod
    async def clear(self) -> None:
        ...

    @abstractmethod
    async def has(self, key: str) -> bool:
        ...

    @abstractmethod
    async def keys(self) -> List[str]:
        ...



from dataclasses import dataclass
from typing import Dict, Literal

import yarl

from graia.amnesia.transport.interface import ExtraContent


@dataclass
class HttpRequest(ExtraContent):
    headers: Dict[str, str]
    cookies: Dict[str, str]
    query_params: Dict[str, str]
    url: yarl.URL
    host: str
    method: 'Literal["GET", "POST", "PUT", "DELETE"] | str'
    client_ip: str
    client_port: int


@dataclass
class HttpResponse(ExtraContent):
    status: int
    headers: Dict[str, str]
    cookies: Dict[str, str]
    uri: yarl.URL



import typing
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, Generic, Iterable, Mapping, Type, TypeVar
from uuid import UUID

T = TypeVar("T")


class _AmnesiaJsonCustomSerialize(Generic[T], Dict[Type[T], Callable[[T], Any]]):
    def new(self, t: Type[T]):
        def decorator(func: Callable[[T], Any]):
            self[t] = func
            return func

        return decorator


SERIALIZERS = _AmnesiaJsonCustomSerialize()

SERIALIZE_UUID_USING_HEX = False
SERIALIZE_DECIMAL_AS_STR = True


def SERIALIZER_DEFAULT(v, d: Dict[Any, Callable[[Any], Any]] = SERIALIZERS):
    return d[v.__class__](v) if v.__class__ in d else v


def DEEP_OBJ_SCAN(x, d: Dict[Any, Callable[[Any], Any]] = SERIALIZERS):
    if isinstance(x, (str, int, float, bool, type(None))):
        return x
    if isinstance(x, Mapping):
        return {k: DEEP_OBJ_SCAN(v, d) for k, v in x.items()}
    if isinstance(x, Iterable):
        return [DEEP_OBJ_SCAN(item, d) for item in x]
    return SERIALIZER_DEFAULT(x)


@SERIALIZERS.new(datetime)
def _(dt: datetime) -> str:
    return dt.isoformat()


@SERIALIZERS.new(timedelta)
def _(td: timedelta) -> str:
    return str(td)


@SERIALIZERS.new(tuple)
def _(t: tuple) -> list:
    return list(t)


@SERIALIZERS.new(set)
def _(s: set) -> list:
    return list(s)


@SERIALIZERS.new(date)
def _(d: date) -> str:
    return d.isoformat()


@SERIALIZERS.new(time)
def _(t: time) -> str:
    return t.isoformat()


@SERIALIZERS.new(Decimal)
def _(d: Decimal):
    return str(d) if SERIALIZE_DECIMAL_AS_STR else float(d)


@SERIALIZERS.new(UUID)
def _(u: UUID) -> str:
    return u.hex if SERIALIZE_UUID_USING_HEX else str(u)


@SERIALIZERS.new(typing.TypedDict)
def _(t: typing.TypedDict) -> dict:
    return dict(t)


@SERIALIZERS.new(typing.NamedTuple)
def _(t: typing.NamedTuple) -> dict:
    return dict(t._asdict())


@SERIALIZERS.new(Enum)
def _(e: Enum) -> str:
    return e.value



from abc import abstractmethod
from typing import Any, Type, TypeVar, Union, overload

from graia.amnesia.transport.common.websocket.operator import (
    WSConnectionAccept,
    WSConnectionClose,
)
from graia.amnesia.transport.exceptions import ConnectionClosed
from graia.amnesia.transport.interface import ExtraContent, PacketIO
from graia.amnesia.transport.signature import TransportSignature

T = TypeVar("T")
E = TypeVar("E", bound=ExtraContent)


class AbstractWebsocketIO(PacketIO[Union[str, bytes]]):
    @abstractmethod
    async def receive(self) -> Union[str, bytes]:
        raise NotImplementedError

    @overload
    @abstractmethod
    async def send(self, data: bytes):
        ...

    @overload
    @abstractmethod
    async def send(self, data: str):
        ...

    @overload
    @abstractmethod
    async def send(self, data: Any):
        ...

    @abstractmethod
    async def send(self, data: ...):
        raise NotImplementedError

    @overload
    @abstractmethod
    async def extra(self, signature: "Type[TransportSignature[T]] | TransportSignature[T]") -> T:
        ...

    @overload
    @abstractmethod
    async def extra(self, signature: "Type[E]") -> E:
        ...

    @abstractmethod
    async def extra(self, signature: Any):
        raise NotImplementedError

    @abstractmethod
    async def wait_for_ready(self):
        pass

    async def accept(self):
        await self.extra(WSConnectionAccept)

    async def close(self):
        if not self.closed:
            await self.extra(WSConnectionClose)

    async def packets(self):
        try:
            while not self.closed:
                yield await self.receive()
        except ConnectionClosed:
            raise
        except Exception:
            pass



from typing import Any, Callable, Coroutine, Union

from graia.amnesia.transport.common.status import ConnectionStatus
from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO
from graia.amnesia.transport.signature import TransportSignature

WebsocketConnectEvent = TransportSignature[Callable[[AbstractWebsocketIO], Coroutine[None, None, Any]]]()
WebsocketReceivedEvent = TransportSignature[
    Callable[[AbstractWebsocketIO, Union[bytes, str]], Coroutine[None, None, Any]]
]()
WebsocketCloseEvent = TransportSignature[Callable[[AbstractWebsocketIO], Coroutine[None, None, Any]]]()
WebsocketReconnect = TransportSignature[Callable[[ConnectionStatus], Coroutine[None, None, bool]]]()



import asyncio
from datetime import timedelta
from heapq import heappop, heappush
from time import time
from typing import Any, Dict, List, Optional, Tuple, Type

from launart import Launart, Service

from graia.amnesia.transport.common.storage import CacheStorage


class Memcache(CacheStorage[Any]):
    cache: Dict[str, Tuple[Optional[float], Any]]
    expire: List[Tuple[float, str]]

    def __init__(
        self,
        cache: Dict[str, Tuple[Optional[float], Any]],
        expire: List[Tuple[float, str]],
    ):
        self.cache = cache
        self.expire = expire

    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        value = self.cache.get(key)
        if value:
            if value[0] is None or value[0] >= time():
                return value[1]
            else:
                del self.cache[key]
        return default

    async def set(self, key: str, value: Any, expire: Optional[timedelta] = None) -> None:
        if expire is None:
            self.cache[key] = (None, value)
            return

        expire_time = time() + expire.total_seconds()
        self.cache[key] = (expire_time, value)
        heappush(self.expire, (expire_time, value))

    async def delete(self, key: str, strict: bool = False) -> None:
        if key in self.cache:
            del self.cache[key]

        elif strict:
            raise KeyError(key)

    async def clear(self) -> None:
        self.cache.clear()
        self.expire.clear()

    async def has(self, key: str) -> bool:
        return key in self.cache

    async def keys(self) -> List[str]:
        return list(self.cache.keys())


class MemcacheService(Service):
    id = "cache.client"
    supported_interface_types = {Memcache}, {CacheStorage: 10}

    interval: float
    cache: Dict[str, Tuple[Optional[float], Any]]
    expire: List[Tuple[float, str]]

    def __init__(
        self,
        interval: float = 0.1,
        cache: Optional[Dict[str, Tuple[Optional[float], Any]]] = None,
        expire: Optional[List[Tuple[float, str]]] = None,
    ):
        self.interval = interval
        self.cache = cache or {}
        self.expire = expire or []
        super().__init__()

    def get_interface(self, interface_type: Type[Memcache]) -> Memcache:
        if issubclass(interface_type, (Memcache)):
            return Memcache(self.cache, self.expire)
        raise ValueError(f"unsupported interface type {interface_type}")

    @property
    def required(self):
        return set()

    @property
    def stages(self):
        return {"blocking"}

    async def launch(self, manager: Launart) -> None:
        async with self.stage("blocking"):
            while not manager.status.exiting:
                if self.expire:
                    expire_time, key = self.expire[0]
                    while expire_time <= time():
                        self.cache.pop(key, None)
                        heappop(self.expire)
                        expire_time, key = self.expire[0]
                await asyncio.sleep(self.interval)



import asyncio
import logging

from launart.manager import Launart
from launart.service import Service
from launart.utilles import wait_fut
from loguru import logger
from uvicorn import Config, Server

from graia.amnesia.builtins.common import ASGIHandlerProvider


class LoguruHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = str(record.levelno)

        frame, depth = logging.currentframe(), 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )


class WithoutSigHandlerServer(Server):
    def install_signal_handlers(self) -> None:
        return


class UvicornService(Service):
    supported_interface_types = set()
    supported_description_types = set()
    id = "http.asgi_runner"
    server: Server
    host: str
    port: int

    def __init__(self, host: str = "127.0.0.1", port: int = 8000):
        self.host = host
        self.port = port
        super().__init__()

    def get_interface(self, interface_type):
        pass

    @property
    def required(self):
        return {"http.universal_server"}

    @property
    def stages(self):
        return {"preparing", "cleanup"}

    async def launch(self, mgr: Launart):
        async with self.stage("preparing"):
            asgi_handler = mgr.get_interface(ASGIHandlerProvider).get_asgi_handler()
            self.server = WithoutSigHandlerServer(Config(asgi_handler, host=self.host, port=self.port))
            # TODO: 使用户拥有更多的对 Config 的配置能力.
            PATCHES = "uvicorn.error", "uvicorn.asgi", "uvicorn.access", ""
            level = logging.getLevelName(20)  # default level for uvicorn
            logging.basicConfig(handlers=[LoguruHandler()], level=level)
            for name in PATCHES:
                target = logging.getLogger(name)
                target.handlers = [LoguruHandler(level=level)]
                target.propagate = False
            serve_task = asyncio.create_task(self.server.serve())
        async with self.stage("cleanup"):
            logger.warning("try to shutdown uvicorn server...")
            self.server.should_exit = True
            await wait_fut([serve_task, asyncio.sleep(10)], return_when=asyncio.FIRST_COMPLETED)
            if not serve_task.done():
                logger.warning("timeout, force exit uvicorn server...")



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from launart import ExportInterface
from launart.service import Service

if TYPE_CHECKING:
    from typing import Any, Awaitable, Callable, MutableMapping, Type


class ASGIHandlerProvider(ExportInterface, metaclass=ABCMeta):
    @abstractmethod
    def get_asgi_handler(
        self,
    ) -> """Callable[[
        Type[MutableMapping[str, Any]],
        Awaitable[MutableMapping[str, Any]],
        Callable[[MutableMapping[str, Any]], Awaitable[None]]
    ], None]""":
        ...


class ExternalASGIHandlerService(Service, ASGIHandlerProvider):
    supported_interface_types = {ASGIHandlerProvider}

    asgi: """Callable[[
        Type[MutableMapping[str, Any]],
        Awaitable[MutableMapping[str, Any]],
        Callable[[MutableMapping[str, Any]], Awaitable[None]]
    ], None]"""

    def __init__(
        self,
        asgi_handler: """Callable[[
        Type[MutableMapping[str, Any]],
        Awaitable[MutableMapping[str, Any]],
        Callable[[MutableMapping[str, Any]], Awaitable[None]]
    ], None]""",
    ) -> None:
        self.asgi = asgi_handler

    def get_interface(self, _):
        return self

    def get_asgi_handler(
        self,
    ) -> """Callable[[
        Type[MutableMapping[str, Any]],
        Awaitable[MutableMapping[str, Any]],
        Callable[[MutableMapping[str, Any]], Awaitable[None]]
    ], None]""":
        return self.asgi



from typing import Any, Optional


class Element:
    def __str__(self) -> str:
        return ""


class Text(Element):
    text: str
    style: Optional[str]

    def __init__(self, text: str, style: Optional[str] = None) -> None:
        """实例化一个 Text 消息元素, 用于承载消息中的文字.

        Args:
            text (str): 元素所包含的文字
            style (Optional[str]): 默认为空, 文字的样式
        """
        self.text = text
        self.style = style

    def __str__(self) -> str:
        return self.text


class Unknown(Element):
    type: str
    raw_data: Any

    def __init__(self, type: str, raw_data: Any) -> None:
        self.type = type
        self.raw_data = raw_data

    def __str__(self) -> str:
        return f"[$Unknown:type={self.type}]"



from abc import abstractmethod
from typing import Any, Dict

from graia.amnesia.transport.common.http.extra import HttpRequest, HttpResponse
from graia.amnesia.transport.interface import ReadonlyIO


class AbstractServerRequestIO(ReadonlyIO[bytes]):
    @abstractmethod
    async def read(self) -> bytes:
        raise NotImplementedError

    @abstractmethod
    async def extra(self, signature: Any):
        raise NotImplementedError

    async def headers(self) -> Dict[str, str]:
        req: HttpRequest = await self.extra(HttpRequest)
        return req.headers

    async def cookies(self) -> Dict[str, str]:
        req: HttpRequest = await self.extra(HttpRequest)
        return req.cookies


class AbstractClientRequestIO(ReadonlyIO[bytes]):
    @abstractmethod
    async def read(self) -> bytes:
        raise NotImplementedError

    @abstractmethod
    async def extra(self, signature: Any):
        raise NotImplementedError

    async def headers(self) -> Dict[str, str]:
        req: HttpResponse = await self.extra(HttpResponse)
        return req.headers

    async def cookies(self) -> Dict[str, str]:
        req: HttpResponse = await self.extra(HttpResponse)
        return req.cookies



from typing import Any, Dict, Optional, Type

import ujson

from graia.amnesia.json import JSONBackend, TJson, TJsonCustomSerializer
from graia.amnesia.json.serializers import DEEP_OBJ_SCAN, SERIALIZERS


class UJsonBackend(JSONBackend):
    def serialize(self, value: Any, *, custom_serializers: Optional[Dict[Type, TJsonCustomSerializer]] = None) -> str:
        return ujson.dumps(
            DEEP_OBJ_SCAN(value, custom_serializers and dict(SERIALIZERS, **custom_serializers) or SERIALIZERS)
        )

    def deserialize(self, value: str) -> TJson:
        return ujson.loads(value)


BACKEND_INSTANCE = UJsonBackend()



from typing import Any

from graia.amnesia.json import TJson

from .bootstrap import CURRENT_BACKEND


class Json:
    @staticmethod
    def serialize_as_bytes(obj: Any) -> bytes:
        return CURRENT_BACKEND.serialize_as_bytes(obj)

    @staticmethod
    def serialize(obj: Any) -> str:
        return CURRENT_BACKEND.serialize(obj)

    @staticmethod
    def deserialize(content: str) -> TJson:
        return CURRENT_BACKEND.deserialize(content)



from functools import partial
from typing import Callable, Dict, List, Optional, TypeVar, cast

from typing_extensions import Concatenate, ParamSpec, Self

from graia.amnesia.transport.signature import TransportSignature

P = ParamSpec("P")
T = TypeVar("T")

# TODO: 连接内简易触发器, 用于通知错误之类的, 或者是 req-res on ws 这种模式上的.
class Transport:
    handlers: Dict[TransportSignature, Callable] = {}
    callbacks: Dict[TransportSignature, List[Callable]] = {}
    declares: List[TransportSignature[None]] = []

    @classmethod
    def __init_subclass__(cls, **kwargs) -> None:
        cls.handlers = {}
        cls.callbacks = {}
        cls.declares = []
        for base in reversed(cls.__bases__):
            if issubclass(base, Transport):
                cls.handlers.update(base.handlers)
                cls.callbacks.update(base.callbacks)
                cls.declares.extend(base.declares)

    def get_handler(self, signature: TransportSignature[Callable[P, T]]) -> Callable[P, T]:
        handler = cast(Optional[Callable[Concatenate[Transport, P], T]], self.handlers.get(signature))
        if handler is None:
            raise TypeError(f"{self.__class__.__name__} has no handler for {signature}")
        return partial(handler, self)  # type: ignore

    def get_callbacks(self, signature: TransportSignature[Callable[P, T]]) -> List[Callable[P, T]]:
        callbacks = cast(Optional[List[Callable[Concatenate[Self, P], T]]], self.callbacks.get(signature))
        if not callbacks:
            raise TypeError(f"{self.__class__.__name__} has no callback for {signature}")
        return [partial(callback, self) for callback in callbacks]  # type: ignore

    def has_handler(self, signature: TransportSignature[Callable]) -> bool:
        return signature in self.handlers

    def has_callback(self, signature: TransportSignature[Callable]) -> bool:
        return signature in self.callbacks

    def iter_handlers(self):
        for signature, unbound in self.handlers.items():
            yield signature, partial(unbound, self)

    def iter_callbacks(self):
        for signature, unbound_callbacks in self.callbacks.items():
            yield signature, [partial(callback, self) for callback in unbound_callbacks]



from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Iterable,
    Iterator,
    List,
    Type,
    TypeVar,
    Union,
)

from typing_extensions import Self

from .element import Element, Text


class MessageChain:
    """即 "消息链", 被用于承载整个消息内容的数据结构, 包含有一有序列表, 包含有继承了 Element 的各式类实例.

    - 你可以通过实例化 `MessageChain` 创建一个消息链

    - `MessageChain.has` 方法可用于判断特定的元素类型是否存在于消息链中

    - `MessageChain.get` 方法可以获取消息链中的所有特定类型的元素

    - `MessageChain.get_first` 方法可以获取消息链中的第 1 个特定类型的元素

    - `MessageChain.get_one` 方法可以获取消息链中的第 index + 1 个特定类型的元素

    - 使用 `str` 函数可以获取到字符串形式表示的消息

    - 使用 `MessageChain(...).join` 方法可以拼接多个消息链并插入指定内容

    - `MessageChain.merge` 方法可以将消息链中相邻的 Text 元素合并为一个 Text 元素.

    - `MessageChain.startswith` 方法可以判断消息链是否以指定的文本开头

    - `MessageChain.endswith` 方法可以判断消息链是否以指定的文本结尾

    - `MessageChain.include` 方法可以创建只包含指定元素类型的消息链

    - `MessageChain.exclude` 方法可以创建排除指定元素类型的消息链

    - `MessageChain.split` 方法可以用指定文本将消息链拆分为多个

    - `MessageChain.append` 方法可以将指定的元素添加到消息链的末尾

    - `MessageChain.extend` 方法可以将指定的序列/消息链添加到消息链的末尾

    - `MessageChain.only` 可以检查是否只包含指定元素

    - `MessageChain.index` 可以获取指定元素在消息链中的索引

    - `MessageChain.count` 可以获取消息链中指定元素的数量

    - `MessageChain.copy` 可以获取消息链的拷贝

    """

    __text_element_class__: ClassVar[Type[Text]] = Text
    content: List[Element]

    def __init__(self, elements: List[Element]):
        """从传入的序列(可以是元组 tuple, 也可以是列表 list) 创建消息链.
        Args:
            elements (List[T]): 包含且仅包含消息元素的序列
        Returns:
            MessageChain: 以传入的序列作为所承载消息的消息链
        """
        self.content = elements

    def has(self, element_class: Type[Element]) -> bool:
        """判断消息链中是否含有特定类型的消息元素
        Args:
            element_class (T): 需要判断的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
        Returns:
            bool: 判断结果
        """
        return element_class in [type(i) for i in self.content]

    if TYPE_CHECKING:
        E = TypeVar("E", bound=Element)

    def get(self, element_class: "Type[E]") -> "List[E]":
        """获取消息链中所有特定类型的消息元素
        Args:
            element_class (T): 指定的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
        Returns:
            List[T]: 获取到的符合要求的所有消息元素; 另: 可能是空列表([]).
        """
        return [i for i in self.content if isinstance(i, element_class)]

    def get_one(self, element_class: "Type[E]", index: int) -> "E":
        """获取消息链中第 index + 1 个特定类型的消息元素
        Args:
            element_class (Type[Element]): 指定的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
            index (int): 索引, 从 0 开始数
        Returns:
            T: 消息链第 index + 1 个特定类型的消息元素
        """
        return self.get(element_class)[index]

    def get_first(self, element_class: "Type[E]") -> "E":
        """获取消息链中第 1 个特定类型的消息元素
        Args:
            element_class (Type[Element]): 指定的消息元素的类型, 例如 "Text", "Notice", "Image" 等.
        Returns:
            T: 消息链第 1 个特定类型的消息元素
        """
        return self.get(element_class)[0]

    def __str__(self) -> str:
        """获取以字符串形式表示的消息链, 且趋于通常你见到的样子.
        Returns:
            str: 以字符串形式表示的消息链
        """
        return "".join(str(i) for i in self.content)

    def join(self, *chains: Union[Self, Iterable[Self]]) -> Self:
        """将多个消息链连接起来, 并在其中插入自身.

        Args:
            *chains (Iterable[MessageChain]): 要连接的消息链.

        Returns:
            MessageChain: 连接后的消息链, 已对文本进行合并.
        """
        result: List[Element] = []
        list_chains: List[MessageChain] = []
        for chain in chains:
            if isinstance(chain, MessageChain):
                list_chains.append(chain)
            else:
                list_chains.extend(chain)

        for chain in list_chains:
            if chain is not list_chains[0]:
                result.extend(deepcopy(self.content))
            result.extend(deepcopy(chain.content))
        return self.__class__(result).merge()

    __contains__ = has

    def __getitem__(self, item: Union[Type[Element], slice]):
        if isinstance(item, slice):
            return self.__class__(self.content[item])
        elif issubclass(item, Element):
            return self.get(item)
        else:
            raise NotImplementedError("{0} is not allowed for item getting".format(type(item)))

    def merge(self) -> Self:
        """合并相邻的 Text 项, 并返回一个新的消息链实例
        Returns:
            MessageChain: 得到的新的消息链实例, 里面不应存在有任何的相邻的 Text 元素.
        """

        result = []

        texts = []
        for i in self.content:
            if not isinstance(i, Text):
                if texts:
                    result.append(self.__class__.__text_element_class__("".join(texts)))
                    texts.clear()  # 清空缓存
                result.append(i)
            else:
                texts.append(i.text)
        if texts:
            result.append(self.__class__.__text_element_class__("".join(texts)))
            texts.clear()  # 清空缓存
        return self.__class__(result)

    def exclude(self, *types: Type[Element]) -> Self:
        """将除了在给出的消息元素类型中符合的消息元素重新包装为一个新的消息链
        Args:
            *types (Type[Element]): 将排除在外的消息元素类型
        Returns:
            MessageChain: 返回的消息链中不包含参数中给出的消息元素类型
        """
        return self.__class__([i for i in self.content if not isinstance(i, types)])

    def include(self, *types: Type[Element]) -> Self:
        """将只在给出的消息元素类型中符合的消息元素重新包装为一个新的消息链
        Args:
            *types (Type[Element]): 将只包含在内的消息元素类型
        Returns:
            MessageChain: 返回的消息链中只包含参数中给出的消息元素类型
        """
        return self.__class__([i for i in self.content if isinstance(i, types)])

    def split(self, pattern: str = " ", raw_string: bool = False) -> List[Self]:
        """和 `str.split` 差不多, 提供一个字符串, 然后返回分割结果.

        Args:
            pattern (str): 分隔符. 默认为单个空格.
            raw_string (bool): 是否要包含 "空" 的文本元素.

        Returns:
            List[Self]: 分割结果, 行为和 `str.split` 差不多.
        """

        result: List[Self] = []
        tmp = []
        for element in self.content:
            if isinstance(element, Text):
                split_result = element.text.split(pattern)
                for index, split_str in enumerate(split_result):
                    if tmp and index > 0:
                        result.append(self.__class__(tmp))
                        tmp = []
                    if split_str or raw_string:
                        tmp.append(self.__class__.__text_element_class__(split_str))
            else:
                tmp.append(element)
        if tmp:
            result.append(self.__class__(tmp))
            tmp = []
        return result

    def __repr__(self) -> str:
        return f"MessageChain({repr(self.content)})"

    def __iter__(self) -> Iterator[Element]:
        yield from self.content

    def startswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串开头

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串开头
        """

        if not self.content or not isinstance(self.content[0], Text):
            return False
        return self.content[0].text.startswith(string)

    def endswith(self, string: str) -> bool:
        """判断消息链是否以给出的字符串结尾

        Args:
            string (str): 字符串

        Returns:
            bool: 是否以给出的字符串结尾
        """

        if not self.content or not isinstance(self.content[-1], Text):
            return False
        return self.content[-1].text.endswith(string)

    def only(self, *element_classes: Type[Element]) -> bool:
        """判断消息链中是否只含有特定类型元素.

        Args:
            *element_classes (Type[Element]): 元素类型

        Returns:
            bool: 判断结果
        """
        return all(isinstance(i, element_classes) for i in self.content)

    def append(self, element: Union[Element, str], copy: bool = False) -> Self:
        """
        向消息链最后追加单个元素

        Args:
            element (Element): 要添加的元素
            copy (bool): 是否要在副本上修改.

        Returns:
            MessageChain: copy = True 时返回副本, 否则返回自己的引用.
        """
        chain_ref = self.copy() if copy else self
        if isinstance(element, str):
            element = self.__class__.__text_element_class__(element)
        chain_ref.content.append(element)
        return chain_ref

    def extend(
        self,
        *content: Union[Self, Element, List[Union[Element, str]]],
        copy: bool = False,
    ) -> Self:
        """
        向消息链最后添加元素/元素列表/消息链

        Args:
            *content (Union[MessageChain, Element, List[Element]]): 要添加的元素/元素容器.
            copy (bool): 是否要在副本上修改.

        Returns:
            MessageChain: copy = True 时返回副本, 否则返回自己的引用.
        """
        result = []
        for i in content:
            if isinstance(i, Element):
                result.append(i)
            elif isinstance(i, str):
                result.append(self.__class__.__text_element_class__(i))
            elif isinstance(i, MessageChain):
                result.extend(i.content)
            else:
                for e in i:
                    if isinstance(e, str):
                        result.append(self.__class__.__text_element_class__(e))
                    else:
                        result.append(e)
        if copy:
            return self.__class__(deepcopy(self.content) + result)
        self.content.extend(result)
        return self

    def copy(self) -> Self:
        """
        拷贝本消息链.

        Returns:
            MessageChain: 拷贝的副本.
        """
        return self.__class__(deepcopy(self.content))

    def index(self, element_type: Type[Element]) -> Union[int, None]:
        """
        寻找第一个特定类型的元素, 并返回其下标.

        Args:
            element_type (Type[Element]): 元素或元素类型

        Returns:
            Optional[int]: 元素下标, 若未找到则为 None.

        """
        return next((i for i, e in enumerate(self.content) if isinstance(e, element_type)), None)

    def count(self, element: Union[Type[Element], Element]) -> int:
        """
        统计共有多少个指定的元素.

        Args:
            element (Type[Element] | Element): 元素或元素类型

        Returns:
            int: 元素数量
        """
        if isinstance(element, Element):
            return sum(i == element for i in self.content)
        return sum(isinstance(i, element) for i in self.content)



from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, List, Type, TypeVar

from typing_extensions import Concatenate, ParamSpec

from graia.amnesia.transport.signature import TransportSignature

if TYPE_CHECKING:
    from graia.amnesia.transport import Transport


P_TransportHandler = ParamSpec("P_TransportHandler")

T_Transport = TypeVar("T_Transport", bound="Transport")


class TransportRegistrar:
    handlers: Dict[TransportSignature, Callable]
    callbacks: Dict[TransportSignature, List]
    declares: List[TransportSignature[None]]

    def __init__(self) -> None:
        self.handlers = {}
        self.callbacks = {}
        self.declares = []

    def handle(self, signature: TransportSignature[Callable[P_TransportHandler, Any]]):
        def decorator(
            method: Callable[Concatenate[T_Transport, P_TransportHandler], Any]
        ) -> Callable[Concatenate[T_Transport, P_TransportHandler], Any]:
            self.handlers[signature] = method
            return method

        return decorator

    def on(self, signature: TransportSignature[Callable[P_TransportHandler, Any]]):
        def decorator(
            method: Callable[Concatenate[T_Transport, P_TransportHandler], Any]
        ) -> Callable[Concatenate[T_Transport, P_TransportHandler], Any]:
            self.callbacks.setdefault(signature, []).append(method)
            return method

        return decorator

    def declare(self, signature: TransportSignature[Any]):
        self.declares.append(signature)
        return signature

    def apply(self, transport_class: Type[T_Transport]) -> Type[T_Transport]:
        transport_class.handlers.update(self.handlers)
        transport_class.callbacks.update(self.callbacks)
        transport_class.declares.extend(self.declares)
        return transport_class



from statv import Stats, Statv


class ConnectionStatus(Statv):
    connected = Stats[bool]("connected", default=False)
    succeed = Stats[bool]("succeed", default=False)

    def __init__(self) -> None:
        super().__init__()

    @property
    def closed(self) -> bool:
        return not self.connected

    @property
    def available(self) -> bool:
        return self.connected



CURRENT_BACKEND = None

try:
    import json
except ImportError:
    pass
else:
    from .backend.std import BACKEND_INSTANCE

    CURRENT_BACKEND = BACKEND_INSTANCE

try:
    import ujson
except ImportError:
    pass
else:
    from .backend.ujson import BACKEND_INSTANCE

    CURRENT_BACKEND = BACKEND_INSTANCE

try:
    import orjson
except ImportError:
    pass
else:
    from .backend.orjson import BACKEND_INSTANCE

    CURRENT_BACKEND = BACKEND_INSTANCE



from dataclasses import dataclass, field
from typing import Callable, Final, Generic, List, Literal, TypeVar, Union

from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO
from graia.amnesia.transport.rider import TransportRider
from graia.amnesia.transport.signature import TransportSignature

from .event import WebsocketCloseEvent as WebsocketCloseEvent
from .event import WebsocketConnectEvent as WebsocketConnectEvent
from .event import WebsocketReceivedEvent as WebsocketReceivedEvent
from .event import WebsocketReconnect as WebsocketReconnect
from .io import AbstractWebsocketIO as AbstractWebsocketIO
from .operator import WSConnectionAccept as WSConnectionAccept
from .operator import WSConnectionClose as WSConnectionClose


@dataclass
class WebsocketEndpoint(
    TransportSignature[None],
):
    path: str
    methods: Final[List[str]] = field(default_factory=lambda: ["GET"])
    method: Final[str] = "WS"

    def __hash__(self) -> int:
        return hash(self.path) + hash(id(self.methods))



import json
from functools import partial
from typing import Any, Dict, Optional, Type

from .. import JSONBackend, TJson, TJsonCustomSerializer
from ..serializers import SERIALIZER_DEFAULT, SERIALIZERS


class StdBackend(JSONBackend):
    def serialize(self, value: Any, *, custom_serializers: Optional[Dict[Type, TJsonCustomSerializer]] = None) -> str:
        return json.dumps(
            value,
            default=partial(
                SERIALIZER_DEFAULT, d=custom_serializers and dict(SERIALIZERS, **custom_serializers) or SERIALIZERS
            ),
        )

    def deserialize(self, value: str) -> TJson:
        return json.loads(value)


BACKEND_INSTANCE = StdBackend()



from typing import Any, Dict, Optional, Tuple, TypedDict, Union

from typing_extensions import NotRequired, Unpack

T_HttpResponse = Union[Tuple[Any, Unpack[Tuple[Dict[str, Any], ...]]], Any]


def status(code: int):
    return {"status": code}


def headers(headers: Dict[str, str]):
    return {"headers": headers}


def cookies(cookies: Dict[str, str], expires: Optional[int] = None):
    return {"cookies": cookies, "cookie_expires": expires}


class HttpServerResponseDescription(TypedDict):
    status: NotRequired[int]
    headers: NotRequired[Dict[str, str]]
    cookies: NotRequired[Dict[str, str]]
    cookie_expires: NotRequired[Optional[int]]


class Response:
    body: Any
    description: HttpServerResponseDescription

    def __init__(self, response_body: Any, *desc: Dict[str, Any]):
        self.body = response_body
        self.description = {}
        for i in desc:
            self.description.update(i)  # type: ignore



from functools import wraps
from typing import TYPE_CHECKING, Any, AnyStr, Callable, Type, TypeVar

from graia.amnesia.json import TJson
from graia.amnesia.json.frontend import Json

if TYPE_CHECKING:
    from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO


_S = TypeVar("_S")
_R = TypeVar("_R")


def data_type(type: Type[AnyStr], error: bool = False):
    def decorator(func: Callable[[Any, "AbstractWebsocketIO", AnyStr], Any]):
        @wraps(func)
        async def wrapper(self, io, data: Any):
            if isinstance(data, type):
                return await func(self, io, data)
            elif error:
                raise TypeError(f"Expected {type.__name__}, got {data.__class__.__name__}")

        return wrapper

    return decorator


def json_require(
    func: Callable[[_S, "AbstractWebsocketIO", TJson], _R]
) -> Callable[[_S, "AbstractWebsocketIO", str], _R]:
    @wraps(func)
    def wrapper(self: _S, io, data: str) -> _R:
        return func(self, io, Json.deserialize(data))

    return wrapper



import asyncio
import pathlib
from functools import partial, reduce
from typing import Optional, Union
from weakref import WeakValueDictionary

import yarl
from launart import ExportInterface, Service
from launart.utilles import wait_fut
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import FileResponse, JSONResponse, PlainTextResponse, Response
from starlette.websockets import WebSocket, WebSocketState

from graia.amnesia.builtins.common import ASGIHandlerProvider
from graia.amnesia.transport import Transport
from graia.amnesia.transport.common.http import AbstractServerRequestIO
from graia.amnesia.transport.common.http import HttpEndpoint as HttpEndpoint
from graia.amnesia.transport.common.http.extra import HttpRequest
from graia.amnesia.transport.common.websocket import AbstractWebsocketIO
from graia.amnesia.transport.common.websocket import (
    WebsocketEndpoint as WebsocketEndpoint,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketCloseEvent as WebsocketCloseEvent,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketConnectEvent as WebsocketConnectEvent,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketReceivedEvent as WebsocketReceivedEvent,
)
from graia.amnesia.transport.common.websocket.operator import (
    WSConnectionAccept as WebsocketAccept,
)
from graia.amnesia.transport.common.websocket.operator import (
    WSConnectionClose as WebsocketClose,
)
from graia.amnesia.transport.exceptions import ConnectionClosed
from graia.amnesia.transport.rider import TransportRider
from graia.amnesia.utilles import random_id


class StarletteServer(ASGIHandlerProvider):
    starlette: Starlette
    service: "StarletteService"

    def __init__(self, service: "StarletteService", starlette: Starlette):
        self.service = service
        self.starlette = starlette

        super().__init__()

    def get_asgi_handler(self):
        return self.starlette


class StarletteRequestIO(AbstractServerRequestIO):
    request: Request

    def __init__(self, request: Request) -> None:
        self.request = request

    async def read(self) -> bytes:
        return await self.request.body()

    async def extra(self, signature):
        if signature is HttpRequest:
            url = self.request.url
            return HttpRequest(
                dict(self.request.headers),
                self.request.cookies,
                dict(self.request.query_params),
                yarl.URL.build(
                    scheme=url.scheme,
                    user=url.username,
                    password=url.password,
                    host=url.hostname or "",
                    port=url.port,
                    path=url.path,
                    query=url.query,
                    fragment=url.fragment,
                ),
                self.request.url.hostname or self.request.headers.get("Host", ""),
                self.request.method,
                self.request.client.host if self.request.client else ".0.0.0",
                self.request.client.port if self.request.client else 0,
            )


class StarletteWebsocketIO(AbstractWebsocketIO):
    websocket: WebSocket
    ready: asyncio.Event

    def __init__(self, websocket: WebSocket) -> None:
        self.websocket = websocket
        self.ready = asyncio.Event()

    async def receive(self) -> Union[str, bytes]:
        received = await self.websocket.receive()
        if "text" in received:
            return received["text"]
        elif "bytes" in received:
            return received["bytes"]
        else:
            if self.websocket.client_state == WebSocketState.DISCONNECTED:
                raise ConnectionClosed("Connection closed")
            raise TypeError(f"Unknown type of received message {received}")

    async def send(self, data: Union[bytes, str]):
        if isinstance(data, str):
            await self.websocket.send_text(data)
        elif isinstance(data, bytes):
            await self.websocket.send_bytes(data)
        else:
            raise TypeError("Unknown type of data to send")

    async def extra(self, signature):
        if signature is HttpRequest:
            url = self.websocket.url
            return HttpRequest(
                dict(self.websocket.headers),
                self.websocket.cookies,
                dict(self.websocket.query_params),
                yarl.URL.build(
                    scheme=url.scheme,
                    user=url.username,
                    password=url.password,
                    host=url.hostname or "",
                    port=url.port,
                    path=url.path,
                    query=url.query,
                    fragment=url.fragment,
                ),
                self.websocket.url.hostname or self.websocket.headers.get("Host", ""),
                "WS",
                self.websocket.client.host if self.websocket.client else "0.0.0.0",
                self.websocket.client.port if self.websocket.client else 0,
            )
        elif signature is WebsocketAccept:
            await self.websocket.accept()
        elif signature is WebsocketClose:
            await self.websocket.close()

    async def headers(self):
        return dict(self.websocket.headers)

    async def cookies(self):
        return self.websocket.cookies

    async def close(self):
        await self.websocket.close()

    async def wait_for_ready(self):
        return await self.ready.wait()

    async def accept(self):
        self.ready.set()
        return await super().accept()

    @property
    def closed(self):
        return self.websocket.application_state == WebSocketState.DISCONNECTED


class StarletteRouter(ExportInterface, TransportRider[str, Union[StarletteRequestIO, StarletteWebsocketIO]]):
    def __init__(self, starlette: Starlette):
        self.starlette = starlette
        self.connections = WeakValueDictionary()
        self.transports = []

    async def http_request_handler(self, handler, request: Request):
        io = StarletteRequestIO(request)
        conn_id = random_id()
        self.connections[conn_id] = io
        try:
            response = await handler(io)
        finally:
            del self.connections[conn_id]

        if not isinstance(response, tuple):
            response_body = response
            response_desc = {}
        else:
            response_body = response[0]
            response_desc = reduce(lambda x, y: dict(x, **y), response[1:], {})

        if isinstance(response_body, (str, bytes)):
            starlette_resp = PlainTextResponse(
                response_body, status_code=response_desc.get("status", 200), headers=response_desc.get("headers", {})
            )
        elif isinstance(response_body, (dict, list)):
            starlette_resp = JSONResponse(
                response_body, status_code=response_desc.get("status", 200), headers=response_desc.get("headers", {})
            )
        elif isinstance(response_body, pathlib.Path):
            starlette_resp = FileResponse(
                response_body, status_code=response_desc.get("status", 200), headers=response_desc.get("headers", {})
            )
        elif isinstance(response_body, Response):
            starlette_resp = response_body
        else:
            raise ValueError(f"unsupported response type {type(response_body)}")

        if "cookies" in response_desc:
            expire = response_desc.get("cookie_expires")
            for key, value in response_desc["cookies"].items():
                starlette_resp.set_cookie(key, value, expire)  # type: ignore

        return starlette_resp

    async def trigger_callbacks(self, event, *args, **kwargs):
        callbacks = [i.get_callbacks(event) for i in self.transports if i.has_callback(event)]
        if callbacks:
            callbacks = reduce(lambda a, b: a + b, callbacks)
            await wait_fut([i(*args, **kwargs) for i in callbacks])

    async def websocket_handler(self, ws: WebSocket):
        io = StarletteWebsocketIO(ws)
        conn_id = random_id()
        self.connections[conn_id] = io
        await self.trigger_callbacks(WebsocketConnectEvent, io)
        if io.closed:
            return
        io.ready.set()
        try:
            async for message in io.packets():
                await self.trigger_callbacks(WebsocketReceivedEvent, io, message)
        except ConnectionClosed:
            await self.trigger_callbacks(WebsocketCloseEvent, io)

    def io(self, id: Optional[str] = None):
        if id is None:
            raise ValueError("id is required")

        return self.connections.get(id)

    def use(self, transport: Transport):
        self.transports.append(transport)

        for signature, handler in transport.iter_handlers():
            if isinstance(signature, HttpEndpoint):
                self.starlette.add_route(
                    signature.path, partial(self.http_request_handler, handler), methods=signature.methods
                )

        for signature in transport.declares:
            if isinstance(signature, WebsocketEndpoint):
                self.starlette.add_websocket_route(signature.path, self.websocket_handler)


class StarletteService(Service):
    id = "http.universal_server"
    supported_interface_types = {ASGIHandlerProvider, StarletteRouter}

    starlette: Starlette

    def __init__(self, starlette: Optional[Starlette] = None) -> None:
        self.starlette = starlette or Starlette()
        super().__init__()

    def get_interface(self, interface_type):
        if issubclass(interface_type, (ASGIHandlerProvider)):
            return StarletteServer(self, self.starlette)
        elif issubclass(interface_type, (StarletteRouter)):
            return StarletteRouter(self.starlette)
        raise ValueError(f"unsupported interface type {interface_type}")

    @property
    def stages(self):
        return set()

    @property
    def required(self):
        return set()

    async def launch(self, _):
        return



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Coroutine, List, Literal, Optional

from launart.service import ExportInterface, TService

from graia.amnesia.json import TJson
from graia.amnesia.transport.common.http.io import AbstractServerRequestIO
from graia.amnesia.transport.common.http.responses import T_HttpResponse
from graia.amnesia.transport.rider import TransportRider
from graia.amnesia.transport.signature import TransportSignature


@dataclass
class HttpEndpoint(
    TransportSignature[
        Callable[
            [AbstractServerRequestIO],
            Coroutine[None, None, T_HttpResponse],
        ]
    ]
):
    path: str
    methods: List[Literal["GET", "POST", "PUT", "DELETE"]] = field(default_factory=lambda: ["GET"])

    def __hash__(self) -> int:
        return hash(self.path) + hash(id(self.methods))


class AbstractClientInterface(ExportInterface[TService], metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, service: TService) -> None:
        raise NotImplementedError

    @abstractmethod
    def request(
        self,
        method: str,
        url: str,
        params: Optional[dict] = None,
        data: Optional[Any] = None,
        headers: Optional[dict] = None,
        cookies: Optional[dict] = None,
        timeout: Optional[float] = None,
        *,
        json: Optional[TJson] = None,
        **kwargs: Any,
    ) -> TransportRider:
        raise NotImplementedError

    def websocket(self, url: str, **kwargs) -> TransportRider:
        raise NotImplementedError



import abc
import inspect
from typing import TYPE_CHECKING, Generic, Type, TypeVar, overload

from graia.amnesia.transport.signature import TransportSignature

if TYPE_CHECKING:
    from graia.amnesia.transport.rider import TransportRider


SPECIALLISTS = {"fields"}


class ExtraContent:
    @classmethod
    def fields(cls):
        return [
            name
            for name, field in inspect.getmembers(cls)
            if not name.startswith("_") and not callable(field) and name not in SPECIALLISTS
        ] + list(cls.__annotations__.keys())


E = TypeVar("E", bound=ExtraContent)
T = TypeVar("T")


class TransportIO(Generic[T], metaclass=abc.ABCMeta):
    rider: "TransportRider"

    @overload
    @abc.abstractmethod
    async def extra(self, signature: "Type[E]") -> E:
        ...

    @overload
    @abc.abstractmethod
    async def extra(self, signature: "Type[TransportSignature[T]] | TransportSignature[T]") -> T:
        ...

    @abc.abstractmethod
    async def extra(self, signature: ...):
        raise NotImplementedError

    def close(self):
        pass

    @property
    def closed(self):
        return False


class PacketIO(TransportIO[T]):
    @abc.abstractmethod
    async def receive(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    async def send(self, data: T):
        raise NotImplementedError


class StreamIO(TransportIO[T]):
    @abc.abstractmethod
    async def read(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    async def write(self, data: T):
        raise NotImplementedError


# 用于描述直接发过来一个然后让你 response 的, 像 HttpResponse 这种, 然后请求信息什么的全让你 extra.
# 然后结果就直接 handler 返回值, 主要用于 http server.
class ReadonlyIO(TransportIO[T]):
    @abc.abstractmethod
    async def read(self) -> T:
        raise NotImplementedError



from __future__ import annotations

import asyncio
import contextlib
import pathlib
from functools import partial, reduce
from io import IOBase
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)
from weakref import WeakValueDictionary

import aiohttp
from aiohttp import (
    ClientResponse,
    ClientSession,
    ClientTimeout,
    ClientWebSocketResponse,
    WSMsgType,
    web,
)
from launart import ExportInterface, Service
from launart.manager import Launart
from launart.utilles import wait_fut
from loguru import logger
from statv import Stats
from typing_extensions import ParamSpec, Self

from graia.amnesia.json import Json, TJson
from graia.amnesia.transport import Transport
from graia.amnesia.transport.common.http import (
    AbstractClientInterface,
    AbstractServerRequestIO,
)
from graia.amnesia.transport.common.http import HttpEndpoint as HttpEndpoint
from graia.amnesia.transport.common.http.extra import HttpRequest, HttpResponse
from graia.amnesia.transport.common.http.io import AbstractClientRequestIO
from graia.amnesia.transport.common.status import ConnectionStatus
from graia.amnesia.transport.common.websocket import AbstractWebsocketIO
from graia.amnesia.transport.common.websocket import (
    WebsocketEndpoint as WebsocketEndpoint,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketCloseEvent as WebsocketCloseEvent,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketConnectEvent as WebsocketConnectEvent,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketReceivedEvent as WebsocketReceivedEvent,
)
from graia.amnesia.transport.common.websocket.event import (
    WebsocketReconnect as WebsocketReconnect,
)
from graia.amnesia.transport.common.websocket.io import AbstractWebsocketIO
from graia.amnesia.transport.common.websocket.operator import (
    WSConnectionAccept as WebsocketAccept,
)
from graia.amnesia.transport.common.websocket.operator import (
    WSConnectionClose as WebsocketClose,
)
from graia.amnesia.transport.exceptions import ConnectionClosed
from graia.amnesia.transport.rider import TransportRider
from graia.amnesia.transport.signature import TransportSignature
from graia.amnesia.utilles import random_id


class AiohttpConnectionStatus(ConnectionStatus):
    drop = Stats[bool]("drop", default=False)

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return f"<ConnectionStatus {self.connected=:}, {self.succeed=:}, {self.drop=:}, {self._waiters=:}>"

    async def wait_for_drop(self) -> None:
        while not self.drop:
            await self.wait_for_update()


class ClientRequestIO(AbstractClientRequestIO):
    rider: ClientConnectionRider[ClientResponse]
    response: ClientResponse

    def __init__(self, rider: ClientConnectionRider) -> None:
        assert rider.response
        self.rider = rider
        self.response = rider.response

    async def read(self) -> bytes:
        data = await self.response.read()
        if self.rider.status.connected and not self.rider.status.drop:
            self.close()
        return data

    async def extra(self, signature):
        if signature is HttpResponse:
            return HttpResponse(
                self.response.status,
                dict(self.response.headers),
                {k: str(v) for k, v in self.response.cookies.items()},
                self.response.url,
            )

    def close(self):
        self.rider.status.drop = True


class ClientWebsocketIO(AbstractWebsocketIO):
    rider: ClientConnectionRider[ClientWebSocketResponse]
    connection: ClientWebSocketResponse

    def __init__(self, rider: ClientConnectionRider) -> None:
        assert rider.response
        self.rider = rider
        self.connection = rider.response

    async def cookies(self) -> Dict[str, str]:
        return {k: v.value for k, v in self.connection._response.cookies.items()}

    async def extra(self, signature):
        if signature is WebsocketClose:
            if not self.connection.closed:
                await self.connection.close()
        elif signature is HttpResponse:
            return HttpResponse(
                self.connection._response.status,
                dict(self.connection._response.request_info.headers),
                {k: v.value for k, v in self.connection._response.cookies.items()},
                self.connection._response.request_info.url,
            )

    async def receive(self) -> Union[bytes, str]:
        msg = await self.connection.receive()
        if msg.type in {WSMsgType.TEXT, WSMsgType.BINARY}:
            return msg.data
        # 错误处理
        if msg.type in {WSMsgType.CLOSE, WSMsgType.CLOSED}:
            raise ConnectionClosed("websocket closed")
        elif msg.type == WSMsgType.ERROR:
            raise msg.data
        else:
            raise TypeError(f"unexpected websocket message type: {msg.type}")

    async def send(self, data: "str | bytes | Any"):
        try:
            if isinstance(data, str):
                await self.connection.send_str(data)
            elif isinstance(data, bytes):
                await self.connection.send_bytes(data)
            else:
                await self.connection.send_json(data)
        except ConnectionError as e:
            raise ConnectionClosed(e.__class__.__qualname__, *e.args) from None

    @property
    def closed(self):
        return self.connection.closed

    async def wait_for_ready(self):
        if self.closed:
            raise ConnectionClosed("websocket closed")


T = TypeVar("T", ClientResponse, ClientWebSocketResponse)

P = ParamSpec("P")


class ClientConnectionRider(TransportRider[str, T], Generic[T]):
    def __init__(
        self,
        interface: AiohttpClientInterface,
        conn_func: Callable[..., AsyncContextManager[T]],
        call_param: Dict[str, Any],
    ) -> None:
        self.transports: List[Transport] = []
        self.connections = {}
        self.response: Optional[T] = None
        self.conn_func = conn_func
        self.call_param = call_param
        self.status = AiohttpConnectionStatus()
        self.autoreceive: bool = False
        self.task = None
        self.connect_task = None
        self.interface = interface

    async def _connect(self):
        async with self.conn_func(**self.call_param) as resp:
            self.response = resp
            self.status.connected = True
            await self.status.wait_for_drop()
            self.status.drop = False
        self.status.connected = False

    async def _start_conn(self) -> Self:
        if not self.status.available:
            self.response = None
            if self.connect_task is None:
                self.connect_task = asyncio.create_task(self._connect())
            await wait_fut((self.status.wait_for_available(), self.connect_task), return_when=asyncio.FIRST_COMPLETED)
            if self.connect_task.done():
                exc = self.connect_task.exception()
                self.connect_task = None
                if exc:
                    raise exc
            self.status.succeed = True
        return self

    def __await__(self):
        return self._start_conn().__await__()

    @overload
    def io(self: "ClientConnectionRider[ClientResponse]") -> ClientRequestIO:
        ...

    @overload
    def io(self: "ClientConnectionRider[ClientWebSocketResponse]") -> ClientWebsocketIO:
        ...

    def io(self, id=None) -> ...:
        if id:
            raise TypeError("this rider has just one connection")
        if self.autoreceive:
            raise TypeError("this rider has been taken over by auto receive, use .use(transport) instead.")
        if not self.status.connected:
            raise RuntimeError("the connection is not ready, please await the instance to ensure connection")
        assert self.response
        if isinstance(self.response, ClientWebSocketResponse):
            return ClientWebsocketIO(self)
        elif isinstance(self.response, ClientResponse):
            return ClientRequestIO(self)
        else:
            raise TypeError("this response is not a ClientResponse or ClientWebSocketResponse")

    async def trigger_callbacks(self, event: TransportSignature[Callable[P, Any]], *args: P.args, **kwargs: P.kwargs):
        tasks: List[asyncio.Task] = []
        for i in self.transports:
            if i.has_callback(event):
                tasks.extend(asyncio.create_task(f(*args, **kwargs)) for f in i.get_callbacks(event))
        for task in tasks:
            await task

    async def connection_manage(self):
        __original_transports: List[Transport] = self.transports[:]

        with contextlib.suppress(Exception):
            while self.transports and self.interface.service.status.stage != "finished":
                try:
                    await self._start_conn()
                    assert isinstance(
                        self.response, ClientWebSocketResponse
                    ), f"{self.response} is not a ClientWebSocketResponse"
                    io = ClientWebsocketIO(self)
                    await self.trigger_callbacks(WebsocketConnectEvent, io)
                    with contextlib.suppress(ConnectionClosed):
                        async for data in io.packets():
                            await self.trigger_callbacks(WebsocketReceivedEvent, io, data)
                    await self.trigger_callbacks(WebsocketCloseEvent, io)
                    if not io.closed:
                        await io.close()
                    self.status.drop = True
                    await self.status.wait_for_unavailable()
                    self.connect_task = None
                except aiohttp.ClientConnectionError as e:
                    logger.warning(repr(e))
                except Exception as e:
                    logger.exception(e)
                # scan transports
                if self.interface.service.status == "cleanup":
                    continue
                continuing_transports: List[Transport] = self.transports[:]
                self.transports = []
                reconnect_handle_tasks = []
                for t in continuing_transports:
                    if t.has_handler(WebsocketReconnect):
                        handler = t.get_handler(WebsocketReconnect)
                        tsk = asyncio.create_task(handler(self.status))
                        tsk.add_done_callback(lambda tsk: self.transports.append(t) if tsk.result() is True else None)
                        reconnect_handle_tasks.append(tsk)
                await wait_fut(reconnect_handle_tasks)
        self.transports = __original_transports

    def use(self, transport: Transport):
        self.autoreceive = True
        self.transports.append(transport)
        if not self.task:
            self.task = asyncio.create_task(self.connection_manage())
        return self.task


class AiohttpClientInterface(AbstractClientInterface["AiohttpService"]):
    service: AiohttpService

    def __init__(self, service: AiohttpService) -> None:
        self.service = service

    def request(
        self,
        method: str,
        url: str,
        params: Optional[dict] = None,
        data: Optional[Any] = None,
        headers: Optional[dict] = None,
        cookies: Optional[dict] = None,
        timeout: Optional[float] = None,
        *,
        json: Optional[TJson] = None,
        **kwargs: Any,
    ) -> ClientConnectionRider[ClientResponse]:
        if json:
            data = Json.serialize(json)
        call_param: Dict[str, Any] = {
            "method": method,
            "url": url,
            "params": params,
            "data": data,
            "headers": headers,
            "cookies": cookies,
            "timeout": timeout,
            **kwargs,
        }
        return ClientConnectionRider(self, self.service.session.request, call_param)

    def websocket(self, url: str, **kwargs) -> ClientConnectionRider[ClientWebSocketResponse]:
        call_param: Dict[str, Any] = {"url": url, **kwargs}
        return ClientConnectionRider(self, self.service.session.ws_connect, call_param)


class AiohttpService(Service):
    id = "http.universal_client"
    session: ClientSession
    supported_interface_types = {AiohttpClientInterface}

    def __init__(self, session: Optional[ClientSession] = None) -> None:
        if TYPE_CHECKING:
            session = session or ClientSession()
        self.session = session
        super().__init__()

    def get_interface(self, interface_type):
        if interface_type is AiohttpClientInterface:
            return AiohttpClientInterface(self)

    @property
    def stages(self):
        return {"preparing", "cleanup"}

    @property
    def required(self):
        return set()

    async def launch(self, mgr: Launart):
        async with self.stage("preparing"):
            if not self.session:
                self.session = ClientSession(timeout=ClientTimeout(total=None))
        async with self.stage("cleanup"):
            await self.session.close()


class AiohttpServerRequestIO(AbstractServerRequestIO):
    request: web.Request

    def __init__(self, request: web.Request) -> None:
        self.request = request

    async def read(self) -> bytes:
        return await self.request.content.read()

    async def extra(self, signature):
        if signature is HttpRequest:
            return HttpRequest(
                dict(self.request.headers),
                dict(self.request.cookies),
                dict(self.request.query),
                self.request.url,
                self.request.host,
                self.request.method,
                self.request.remote or ".0.0.0",
                0,
            )


class AiohttpServerWebsocketIO(AbstractWebsocketIO):
    websocket: web.WebSocketResponse
    request: web.Request
    ready: asyncio.Event

    def __init__(self, request: web.Request) -> None:
        self.websocket = web.WebSocketResponse(autoping=True)
        self.request = request
        self.ready = asyncio.Event()

    async def receive(self) -> Union[str, bytes]:
        try:
            received = await self.websocket.receive()
        except asyncio.CancelledError as e:
            raise ConnectionClosed("Cancelled") from e
        if received.type in (web.WSMsgType.BINARY, web.WSMsgType.TEXT):
            return received.data
        elif received.type in (web.WSMsgType.CLOSE, web.WSMsgType.CLOSING, web.WSMsgType.CLOSED):
            self.ready.clear()
            raise ConnectionClosed("Connection closed")
        elif received.type is web.WSMsgType.ERROR:
            exc = self.websocket.exception()
            raise ConnectionClosed("Websocket Error") from exc
        raise TypeError(f"Unknown type of received message {received}")

    async def send(self, data: Union[str, bytes, Any]):
        if isinstance(data, str):
            await self.websocket.send_str(data)
        elif isinstance(data, bytes):
            await self.websocket.send_bytes(data)
        else:
            await self.websocket.send_json(data)

    async def extra(self, signature):
        if signature is HttpRequest:
            return HttpRequest(
                dict(self.request.headers),
                dict(self.request.cookies),
                dict(self.request.query),
                self.request.url,
                self.request.host,
                self.request.method,
                self.request.remote or ".0.0.0",
                0,
            )
        elif signature is WebsocketAccept:
            await self.websocket.prepare(self.request)
            self.ready.set()
        elif signature is WebsocketClose:
            await self.websocket.close()

    async def wait_for_ready(self):
        return await self.ready.wait()

    async def accept(self):
        return await super().accept()

    async def headers(self) -> Dict[str, str]:
        return dict(self.websocket.headers)

    async def cookies(self) -> Dict[str, str]:
        return dict(self.request.cookies)

    async def close(self):
        await self.websocket.close()

    @property
    def closed(self):
        return self.websocket.closed


class AiohttpRouter(
    ExportInterface["AiohttpServerService"],
    TransportRider[str, Union[AiohttpServerRequestIO, AiohttpServerWebsocketIO]],
):
    def __init__(self, wsgi: web.Application) -> None:
        self.connections = WeakValueDictionary()
        self.transports = []
        self.wsgi = wsgi

    async def handle_http_request(self, handler: Callable[[AiohttpServerRequestIO], Any], request: web.Request):
        req_io = AiohttpServerRequestIO(request)
        conn_id = random_id()
        self.connections[conn_id] = req_io
        try:
            handler_resp = await handler(req_io)
        finally:
            del self.connections[conn_id]
        if not isinstance(handler_resp, tuple):
            body = handler_resp
            resp_info = {}
        else:
            body = handler_resp[0]
            resp_info = reduce(lambda x, y: dict(x, **y), handler_resp[1:], {})
        status: int = resp_info.get("status", 200)
        if isinstance(body, (str, bytes, IOBase)):
            response = web.Response(body=body, status=status)
        elif isinstance(body, (dict, list)):
            response = web.json_response(body, status=status, headers=resp_info.get("headers"))
        elif isinstance(body, pathlib.Path):
            response = web.Response(body=body.open("rb"), status=status)
        elif isinstance(body, web.Response):
            response = body
        else:
            raise ValueError(f"unsupported response type {type(body)}")

        if "cookies" in resp_info:
            expire = resp_info.get("cookie_expires")
            for key, value in resp_info["cookies"].items():
                response.set_cookie(key, value, expires=expire)

        return response

    async def handle_websocket_request(self, request: web.Request) -> web.WebSocketResponse:
        websocket_io = AiohttpServerWebsocketIO(request)
        conn_id = random_id()
        self.connections[conn_id] = websocket_io
        await self.trigger_callbacks(WebsocketConnectEvent, websocket_io)
        if websocket_io.closed:
            return websocket_io.websocket
        with contextlib.suppress(ConnectionClosed):
            async for message in websocket_io.packets():
                await self.trigger_callbacks(WebsocketReceivedEvent, websocket_io, message)
        await self.trigger_callbacks(WebsocketCloseEvent, websocket_io)
        await websocket_io.close()
        return websocket_io.websocket

    async def trigger_callbacks(self, event, *args, **kwargs):
        callbacks = [i.get_callbacks(event) for i in self.transports if i.has_callback(event)]
        if callbacks:
            callbacks = reduce(lambda a, b: a + b, callbacks)
            await wait_fut([i(*args, **kwargs) for i in callbacks])

    def io(self, id: Optional[str] = None):
        if id is None:
            raise ValueError("id is required")

        return self.connections.get(id)

    def use(self, transport: Transport):
        self.transports.append(transport)
        for signature, handler in transport.iter_handlers():
            if isinstance(signature, HttpEndpoint):
                for method in signature.methods:
                    self.wsgi.router.add_route(method, signature.path, partial(self.handle_http_request, handler))

        for signature in transport.declares:
            if isinstance(signature, WebsocketEndpoint):
                self.wsgi.router.add_route("GET", signature.path, self.handle_websocket_request)


class AiohttpServerService(Service):
    id = "http.universal_server"
    wsgi_handler: web.Application
    supported_interface_types = {AiohttpRouter}

    def __init__(
        self, host: str = "127.0.0.1", port: int = 8000, wsgi_handler: Optional[web.Application] = None
    ) -> None:
        self.wsgi_handler = wsgi_handler or web.Application()
        self.wsgi_handler.router.freeze = lambda: None  # monkey patch
        self.routers: List[AiohttpRouter] = []
        self.host = host
        self.port = port
        super().__init__()

    def get_interface(self, interface_type):
        if interface_type is AiohttpRouter:
            router = AiohttpRouter(self.wsgi_handler)
            self.routers.append(router)
            return router

    @property
    def stages(self):
        return {"preparing", "blocking", "cleanup"}

    @property
    def required(self):
        return set()

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            logger.info(f"starting server on {self.host}:{self.port}")
            runner = web.AppRunner(self.wsgi_handler)
            await runner.setup()
            site = web.TCPSite(runner, self.host, self.port)
        async with self.stage("blocking"):
            await site.start()
        async with self.stage("cleanup"):
            await self.wsgi_handler.shutdown()
            await self.wsgi_handler.cleanup()



import abc
from typing import Generic, List, MutableMapping, Optional, Type, TypeVar, overload

from graia.amnesia.transport import Transport

K = TypeVar("K")
V = TypeVar("V")


class TransportRider(Generic[K, V], metaclass=abc.ABCMeta):
    connections: MutableMapping[K, V]
    transports: List[Transport]

    @abc.abstractmethod
    def io(self, id: Optional[K] = None):
        raise NotImplementedError

    @abc.abstractmethod
    def use(self, transport: Transport):
        raise NotImplementedError
